/*
 * Copyright (c) 2010-2011, The MiCode Open Source Community (www.micode.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// 定义包名
package net.micode.notes.gtask.data;

// 导入所需的类
import android.content.ContentResolver; // 用于内容解析的类
import android.content.ContentUris; // 处理内容URI的类
import android.content.ContentValues; // 用于存储内容数据的类
import android.content.Context; // Android上下文类
import android.database.Cursor; // 数据库操作的游标类
import android.net.Uri; // 数据库URI的类
import android.util.Log; // 日志记录类

// 导入Note和相关数据列类
import net.micode.notes.data.Notes;
import net.micode.notes.data.Notes.DataColumns;
import net.micode.notes.data.Notes.DataConstants;
import net.micode.notes.data.Notes.NoteColumns;
import net.micode.notes.data.NotesDatabaseHelper.TABLE; // 数据库表
import net.micode.notes.gtask.exception.ActionFailureException; // 自定义异常类

// 导入JSON相关类
import org.json.JSONException;
import org.json.JSONObject;

// SqlData类，用于处理与数据库中的笔记数据交互
public class SqlData {
    // 定义日志Tag
    private static final String TAG = SqlData.class.getSimpleName();

    // 定义无效ID常量
    private static final int INVALID_ID = -99999;

    // 定义用于查询的数据列投影
    public static final String[] PROJECTION_DATA = new String[] {
            DataColumns.ID, DataColumns.MIME_TYPE, DataColumns.CONTENT, DataColumns.DATA1,
            DataColumns.DATA3
    };

    // 定义数据列在投影中的索引
    public static final int DATA_ID_COLUMN = 0; // ID列索引
    public static final int DATA_MIME_TYPE_COLUMN = 1; // MIME类型列索引
    public static final int DATA_CONTENT_COLUMN = 2; // 内容列索引
    public static final int DATA_CONTENT_DATA_1_COLUMN = 3; // 数据1列索引
    public static final int DATA_CONTENT_DATA_3_COLUMN = 4; // 数据3列索引

    // 定义成员变量
    private ContentResolver mContentResolver; // 内容解析器
    private boolean mIsCreate; // 是否为创建状态
    private long mDataId; // 数据ID
    private String mDataMimeType; // 数据MIME类型
    private String mDataContent; // 数据内容
    private long mDataContentData1; // 内容数据1
    private String mDataContentData3; // 内容数据3
    private ContentValues mDiffDataValues; // 存储差异数据的ContentValues

    // 构造函数，用于创建新SQL数据
    public SqlData(Context context) {
        mContentResolver = context.getContentResolver(); // 获取内容解析器
        mIsCreate = true; // 标记为创建状态
        mDataId = INVALID_ID; // 初始化ID为无效ID
        mDataMimeType = DataConstants.NOTE; // 设置默认MIME类型
        mDataContent = ""; // 初始化内容为空字符串
        mDataContentData1 = 0; // 初始化数据1为0
        mDataContentData3 = ""; // 初始化数据3为空字符串
        mDiffDataValues = new ContentValues(); // 创建新的ContentValues
    }

    // 构造函数，用于从Cursor中加载现有SQL数据
    public SqlData(Context context, Cursor c) {
        mContentResolver = context.getContentResolver(); // 获取内容解析器
        mIsCreate = false; // 标记为非创建状态
        loadFromCursor(c); // 从Cursor加载数据
        mDiffDataValues = new ContentValues(); // 创建新的ContentValues
    }

    // 从Cursor加载数据
    private void loadFromCursor(Cursor c) {
        mDataId = c.getLong(DATA_ID_COLUMN); // 获取ID
        mDataMimeType = c.getString(DATA_MIME_TYPE_COLUMN); // 获取MIME类型
        mDataContent = c.getString(DATA_CONTENT_COLUMN); // 获取内容
        mDataContentData1 = c.getLong(DATA_CONTENT_DATA_1_COLUMN); // 获取数据1
        mDataContentData3 = c.getString(DATA_CONTENT_DATA_3_COLUMN); // 获取数据3
    }

    // 设置内容，从JSON中提取数据
    public void setContent(JSONObject js) throws JSONException {
        long dataId = js.has(DataColumns.ID) ? js.getLong(DataColumns.ID) : INVALID_ID; // 获取ID
        if (mIsCreate || mDataId != dataId) {
            mDiffDataValues.put(DataColumns.ID, dataId); // 存储ID到差异数据
        }
        mDataId = dataId; // 更新ID

        String dataMimeType = js.has(DataColumns.MIME_TYPE) ? js.getString(DataColumns.MIME_TYPE)
                : DataConstants.NOTE; // 获取MIME类型
        if (mIsCreate || !mDataMimeType.equals(dataMimeType)) {
            mDiffDataValues.put(DataColumns.MIME_TYPE, dataMimeType); // 存储MIME类型
        }
        mDataMimeType = dataMimeType; // 更新MIME类型

        String dataContent = js.has(DataColumns.CONTENT) ? js.getString(DataColumns.CONTENT) : ""; // 获取内容
        if (mIsCreate || !mDataContent.equals(dataContent)) {
            mDiffDataValues.put(DataColumns.CONTENT, dataContent); // 存储内容
        }
        mDataContent = dataContent; // 更新内容

        long dataContentData1 = js.has(DataColumns.DATA1) ? js.getLong(DataColumns.DATA1) : 0; // 获取数据1
        if (mIsCreate || mDataContentData1 != dataContentData1) {
            mDiffDataValues.put(DataColumns.DATA1, dataContentData1); // 存储数据1
        }
        mDataContentData1 = dataContentData1; // 更新数据1

        String dataContentData3 = js.has(DataColumns.DATA3) ? js.getString(DataColumns.DATA3) : ""; // 获取数据3
        if (mIsCreate || !mDataContentData3.equals(dataContentData3)) {
            mDiffDataValues.put(DataColumns.DATA3, dataContentData3); // 存储数据3
        }
        mDataContentData3 = dataContentData3; // 更新数据3
    }

    // 获取内容，返回JSON对象
    public JSONObject getContent() throws JSONException {
        if (mIsCreate) {
            Log.e(TAG, "it seems that we haven't created this in database yet"); // 日志输出
            return null; // 如果是创建状态，返回null
        }
        JSONObject js = new JSONObject(); // 创建JSON对象
        js.put(DataColumns.ID, mDataId); // 设置ID
        js.put(DataColumns.MIME_TYPE, mDataMimeType); // 设置MIME类型
        js.put(DataColumns.CONTENT, mDataContent); // 设置内容
        js.put(DataColumns.DATA1, mDataContentData1); // 设置数据1
        js.put(DataColumns.DATA3, mDataContentData3); // 设置数据3
        return js; // 返回JSON对象
    }

    // 提交更改到数据库
    public void commit(long noteId, boolean validateVersion, long version) {
        // 如果是创建状态
        if (mIsCreate) {
            // 如果ID为无效ID且差异数据包含ID，则移除ID
            if (mDataId == INVALID_ID && mDiffDataValues.containsKey(DataColumns.ID)) {
                mDiffDataValues.remove(DataColumns.ID);
            }

            mDiffDataValues.put(DataColumns.NOTE_ID, noteId); // 存储笔记ID
            Uri uri = mContentResolver.insert(Notes.CONTENT_DATA_URI, mDiffDataValues); // 插入数据
            try {
                mDataId = Long.valueOf(uri.getPathSegments().get(1)); // 获取新插入数据的ID
            } catch (NumberFormatException e) {
                Log.e(TAG, "Get note id error :" + e.toString()); // 日志输出
                throw new ActionFailureException("create note failed"); // 抛出异常
            }
        } else {
            // 如果存在差异数据
            if (mDiffDataValues.size() > 0) {
                int result = 0; // 结果计数器
                // 判断是否需要验证版本
                if (!validateVersion) {
                    result = mContentResolver.update(ContentUris.withAppendedId(
                            Notes.CONTENT_DATA_URI, mDataId), mDiffDataValues, null, null); // 更新数据
                } else {
                    // 根据版本验证更新数据
                    result = mContentResolver.update(ContentUris.withAppendedId(
                            Notes.CONTENT_DATA_URI, mDataId), mDiffDataValues,
                            " ? in (SELECT " + NoteColumns.ID + " FROM " + TABLE.NOTE
                                    + " WHERE " + NoteColumns.VERSION + "=?)", new String[] {
                                    String.valueOf(noteId), String.valueOf(version)
                            });
                }
                if (result == 0) {
                    Log.w(TAG, "there is no update. maybe user updates note when syncing"); // 日志输出
                }
            }
        }
        mDiffDataValues.clear(); // 清空差异数据
        mIsCreate = false; // 设置为非创建状态
    }

    // 获取数据ID
    public long getId() {
        return mDataId; // 返回数据ID
    }
}
